from uproot3_methods.classes.TH1 import Methods as TH1Methods

# https://github.com/CoffeaTeam/coffea/pull/329


class TH1(TH1Methods, list):
    pass


class TAxis(object):
    def __init__(self, fNbins, fXmin, fXmax):
        self._fNbins = fNbins
        self._fXmin = fXmin
        self._fXmax = fXmax


def export1d_coffea(hist):
    """Export a 1-dimensional `Hist` object to uproot3
    This allows one to write a coffea histogram into a ROOT file, via uproot3.
    Parameters
    ----------
        hist : Hist
            A 1-dimensional histogram object
    Returns
    -------
        out
            A ``uproot3_methods.classes.TH1`` object
    Examples
    --------
    Creating a coffea histogram, filling, and writing to a file::
        import coffea, uproot3, numpy
        h = coffea.hist.Hist("Events", coffea.hist.Bin("var", "some variable", 20, 0, 1))
        h.fill(var=numpy.random.normal(size=100))
        fout = uproot3.create('output.root')
        fout['myhist'] = coffea.hist.export1d(h)
        fout.close()
    """
    from coffea.hist.hist_tools import SparseAxis, DenseAxis

    if hist.dense_dim() != 1:
        raise ValueError("export1d() can only support one dense dimension")
    if hist.sparse_dim() != 0:
        raise ValueError("export1d() expects zero sparse dimensions")

    axis = hist.axes()[0]
    sumw, sumw2 = hist.values(sumw2=True, overflow="all")[()]
    edges = axis.edges(overflow="none")

    out = TH1.__new__(TH1)
    out._fXaxis = TAxis(len(edges) - 1, edges[0], edges[-1])
    out._fXaxis._fName = axis.name
    out._fXaxis._fTitle = axis.label
    if not axis._uniform:
        out._fXaxis._fXbins = edges.astype(">f8")

    centers = (edges[:-1] + edges[1:]) / 2.0
    out._fEntries = out._fTsumw = sumw[1:-1].sum()
    out._fTsumwx = (sumw[1:-1] * centers).sum()
    out._fTsumwx2 = (sumw[1:-1] * centers ** 2).sum()

    out._fName = "histogram"
    out._fTitle = hist.label

    out._classname = b"TH1F"
    out.extend(sumw.astype(">f8"))
    out._fSumw2 = sumw2.astype(">f8")
    out._fTsumw2 = sumw2[1:-1].sum()

    return out


def export1d_hist(h):
    """Export a 1-dimensional `Hist` object to uproot3
    This allows one to write a coffea histogram into a ROOT file, via uproot3.
    Parameters
    ----------
        hist : Hist
            A 1-dimensional histogram object
    Returns
    -------
        out
            A ``uproot3_methods.classes.TH1`` object
    Examples
    --------
    Creating a coffea histogram, filling, and writing to a file::
        import coffea, uproot3, numpy
        h = hist.Hist(hist.axis.Regular(20, 0, 1, name="var"))
        h.fill(var=numpy.random.normal(size=100))
        fout = uproot3.create('output.root')
        fout['myhist'] = export1d_hist(h)
        fout.close()
    """
    import hist

    (axis,) = h.axes
    assert isinstance(axis, (hist.axis.Regular, hist.axis.Variable))

    sumw = h.view(flow=True)["value"]
    sumw2 = h.view(flow=True)["variance"]
    edges = axis.edges

    out = TH1.__new__(TH1)
    out._fXaxis = TAxis(len(edges) - 1, edges[0], edges[-1])
    out._fXaxis._fName = axis.name
    out._fXaxis._fTitle = axis.label
    if not isinstance(axis, hist.axis.Regular):
        out._fXaxis._fXbins = edges.astype(">f8")

    centers = axis.centers
    out._fEntries = out._fTsumw = sumw[1:-1].sum()
    out._fTsumwx = (sumw[1:-1] * centers).sum()
    out._fTsumwx2 = (sumw[1:-1] * centers ** 2).sum()

    out._fName = "histogram"
    out._fTitle = axis.label

    out._classname = b"TH1F"
    out.extend(sumw.astype(">f8"))
    out._fSumw2 = sumw2.astype(">f8")
    out._fTsumw2 = sumw2[1:-1].sum()

    return out
